# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Any

import pytest

from sciline.scheduler import DaskScheduler, NaiveScheduler, Scheduler


@pytest.fixture
def naive_scheduler() -> NaiveScheduler:
    return NaiveScheduler()


@pytest.fixture
def dask_scheduler() -> DaskScheduler | None:
    try:
        import dask  # noqa: F401

        return DaskScheduler()
    except ModuleNotFoundError:
        pytest.skip("Test requires Dask")


@pytest.fixture(scope='session')
def dask_distributed_client() -> Any:
    """Manage a singleton Dask client for all tests.

    Tests must always use this client if they need to use dask.distributed.
    They must never create their own client.
    Otherwise, the clients will be in conflict.
    """
    try:
        from dask.distributed import Client

        return Client(set_as_default=False)
    except ImportError:
        pytest.skip("Test requires dask.distributed")


@pytest.fixture
def dask_distributed_scheduler(dask_distributed_client: Any) -> DaskScheduler | None:
    if dask_distributed_client is None:
        return None
    return DaskScheduler(dask_distributed_client.get)


@pytest.fixture(params=['naive', 'dask', 'dask_distributed'])
def scheduler(request: pytest.FixtureRequest) -> Scheduler:
    match request.param:
        case 'naive':
            return request.getfixturevalue('naive_scheduler')  # type: ignore[no-any-return]
        case 'dask':
            return request.getfixturevalue('dask_scheduler')  # type: ignore[no-any-return]
        case 'dask_distributed':
            return request.getfixturevalue('dask_distributed_scheduler')  # type: ignore[no-any-return]
        case _:
            raise ValueError(f"Unknown scheduler: {request.param}")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from typing import NewType

import pandas as pd
import pytest
from cyclebane.graph import IndexValues, NodeName

import sciline as sl

A = NewType('A', int)
B = NewType('B', int)
C = NewType('C', int)
D = NewType('D', int)
X = NewType('X', int)


def a_to_b(a: A) -> B:
    return B(a + 1)


def b_to_c(b: B) -> C:
    return C(b + 2)


def c_to_d(c: C) -> D:
    return D(c + 4)


def test_map_returns_pipeline_that_can_compute_for_each_value() -> None:
    ab = sl.Pipeline((a_to_b,))
    mapped = ab.map({A: [A(10 * i) for i in range(3)]})
    with pytest.raises(sl.UnsatisfiedRequirement):
        # B is not in the graph any more, since it has been duplicated
        mapped.compute(B)

    for i in range(3):
        index = IndexValues(('dim_0',), (i,))
        assert mapped.compute(NodeName(A, index)) == A(10 * i)  # type: ignore[call-overload]
        assert mapped.compute(NodeName(B, index)) == B(10 * i + 1)  # type: ignore[call-overload]


def test_reduce_returns_pipeline_passing_mapped_branches_to_reducing_func() -> None:
    ab = sl.Pipeline((a_to_b,))
    mapped = ab.map({A: [A(10 * i) for i in range(3)]})
    # Define key to make mypy happy. This can actually be anything but currently
    # the type-hinting of Pipeline is too specific, disallowing, e.g., strings.
    Result = NewType('Result', int)
    assert mapped.reduce(func=min, name=Result).compute(Result) == Result(1)
    assert mapped.reduce(func=max, name=Result).compute(Result) == Result(21)


def test_compute_mapped_single_index() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    pl[B] = B(7)
    paramsA = pd.DataFrame(
        {A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c']
    ).rename_axis('x')
    mapped = pl.map(paramsA)
    result = sl.compute_mapped(mapped, C)
    assert result['a'] == C(7)
    assert result['b'] == C(17)
    assert result['c'] == C(27)
    assert result.index.name == 'x'
    assert result.name == C


def test_compute_mapped_single_index_with_no_name() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    pl[B] = B(7)
    paramsA = pd.DataFrame({A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c'])
    mapped = pl.map(paramsA)
    result = sl.compute_mapped(mapped, C)
    assert result['a'] == C(7)
    assert result['b'] == C(17)
    assert result['c'] == C(27)
    assert result.index.name == 'dim_0'
    assert result.name == C


def test_compute_mapped_from_mapped_with_implicit_index() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    pl[B] = B(7)
    mapped = pl.map({A: [A(10 * i) for i in range(3)]})
    result = sl.compute_mapped(mapped, C)
    assert result[0] == C(7)
    assert result[1] == C(17)
    assert result[2] == C(27)
    assert result.index.name == 'dim_0'
    assert result.name == C


def test_compute_mapped_multiple_indices_creates_multiindex() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    paramsA = pd.DataFrame(
        {A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c']
    ).rename_axis('x')
    paramsB = pd.DataFrame(
        {B: [B(i) for i in range(2)]}, index=['aa', 'bb']
    ).rename_axis('y')
    mapped = pl.map(paramsA).map(paramsB)
    result = sl.compute_mapped(mapped, C)
    assert result['a', 'aa'] == C(0)
    assert result['a', 'bb'] == C(1)
    assert result['b', 'aa'] == C(10)
    assert result['b', 'bb'] == C(11)
    assert result['c', 'aa'] == C(20)
    assert result['c', 'bb'] == C(21)
    assert result.index.names == list(mapped.index_names)
    assert result.name == C


def test_compute_mapped_ignores_unrelated_index() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    paramsA = pd.DataFrame(
        {A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c']
    ).rename_axis('x')
    paramsB = pd.DataFrame(
        {B: [B(i) for i in range(2)]}, index=['aa', 'bb']
    ).rename_axis('y')
    mapped = pl.map(paramsA).map(paramsB)
    result = sl.compute_mapped(mapped, A)
    assert result.index.name == 'x'
    assert result['a'] == A(0)
    assert result['b'] == A(10)
    assert result['c'] == A(20)
    assert result.name == A


def test_compute_mapped_raises_if_node_is_not_mapped() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    pl[B] = B(7)
    mapped = pl.map({A: [A(10 * i) for i in range(3)]})
    with pytest.raises(ValueError, match='is not a mapped node'):
        sl.compute_mapped(mapped, B)


def test_compute_mapped_raises_if_node_depends_on_but_is_not_mapped() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    pl[B] = B(7)
    pl = pl.map({A: [A(10 * i) for i in range(3)]}).reduce(func=max, name=C)
    pl.insert(c_to_d)
    # Slightly different failure case from above: The relevant subgraph *does* have
    # indices, but the node does not.
    with pytest.raises(ValueError, match='is not a mapped node'):
        sl.compute_mapped(pl, D)


def test_can_compute_subset_of_get_mapped_node_names() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    pl[B] = B(7)
    paramsA = pd.DataFrame(
        {A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c']
    ).rename_axis('x')
    mapped = pl.map(paramsA)
    result = sl.get_mapped_node_names(mapped, C)
    # We lose the convenience of compute_mapped which returns a nicely setup series
    # but this is perfectly fine and possible.
    assert mapped.compute(result.iloc[1]) == A(17)
    assert mapped.compute(result.loc['b']) == A(17)


def test_compute_mapped_raises_if_multiple_mapped_nodes_with_given_name() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    paramsA = pd.DataFrame(
        {A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c']
    ).rename_axis('x')
    paramsB = pd.DataFrame(
        {B: [B(i) for i in range(2)]}, index=['aa', 'bb']
    ).rename_axis('y')
    pl = pl.map(paramsA).map(paramsB).reduce(func=max, name=C, index='x')
    with pytest.raises(ValueError, match='Multiple mapped nodes with name'):
        sl.compute_mapped(pl, C)


def test_compute_mapped_with_partial_reduction_identifies_correct_index() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    D = NewType('D', int)

    pl = sl.Pipeline((ab_to_c,))
    paramsA = pd.DataFrame(
        {A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c']
    ).rename_axis('x')
    paramsB = pd.DataFrame(
        {B: [B(i) for i in range(2)]}, index=['aa', 'bb']
    ).rename_axis('y')
    pl = pl.map(paramsA).map(paramsB).reduce(func=max, name=D, index='x')
    result = sl.compute_mapped(pl, D)
    assert result['aa'] == C(20)
    assert result['bb'] == C(21)


def test_compute_mapped_index_names_selects_between_multiple_candidates() -> None:
    def ab_to_c(a: A, b: B) -> C:
        return C(a + b)

    pl = sl.Pipeline((ab_to_c,))
    paramsA = pd.DataFrame(
        {A: [A(10 * i) for i in range(3)]}, index=['a', 'b', 'c']
    ).rename_axis('x')
    paramsB = pd.DataFrame(
        {B: [B(i) for i in range(2)]}, index=['aa', 'bb']
    ).rename_axis('y')
    mapped = pl.map(paramsA).map(paramsB)
    pl = mapped.reduce(func=max, name=C, index='x')
    result_y = sl.compute_mapped(pl, C, index_names=('y',))
    assert result_y['aa'] == C(20)
    assert result_y['bb'] == C(21)
    for index_names in [('x', 'y'), ('y', 'x')]:
        result_xy = sl.compute_mapped(pl, C, index_names=index_names)
        assert result_xy['a', 'aa'] == C(0)
        assert result_xy['a', 'bb'] == C(1)
        assert result_xy['b', 'aa'] == C(10)
        assert result_xy['b', 'bb'] == C(11)
        assert result_xy['c', 'aa'] == C(20)
        assert result_xy['c', 'bb'] == C(21)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import Callable, Hashable, Iterable, Mapping, Sequence
from itertools import chain
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeVar,
    get_args,
    get_type_hints,
    overload,
)

import networkx as nx
from networkx.algorithms.simple_paths import all_simple_paths

from ._provider import Provider, ToProvider
from ._utils import key_name
from .data_graph import DataGraph, to_task_graph
from .display import pipeline_html_repr
from .handler import ErrorHandler, HandleAsComputeTimeException, UnsatisfiedRequirement
from .reporter import Reporter
from .scheduler import Scheduler
from .task_graph import TaskGraph
from .typing import Key

if TYPE_CHECKING:
    import graphviz
    import pandas


T = TypeVar('T')
KeyType = TypeVar('KeyType', bound=Key)


def _is_multiple_keys(keys: type | Iterable[type] | UnionType) -> bool:
    # Cannot simply use isinstance(keys, Iterable) because that is True for
    # generic aliases of iterable types, e.g.,
    #
    # class Str(sl.Scope[Param, str], str): ...
    # keys = Str[int]
    #
    # And isinstance(keys, type) does not work on its own because
    # it is False for the above type.
    if isinstance(keys, str):
        return False
    return (
        not isinstance(keys, type) and not get_args(keys) and isinstance(keys, Iterable)
    )


def _find_paths_to_targets(
    graph: Any, missing: Any, targets: tuple[Any, ...]
) -> list[list[Any]]:
    """Find paths from missing node to each target.

    Parameters
    ----------
    graph:
        The graph to traverse
    missing:
        The missing node
    targets:
        The target nodes

    Returns
    -------
    :
        List of paths, where each path is a list of nodes from missing to target
    """

    paths = []
    for target in targets:
        try:
            # Find all simple paths from the missing node to the target
            paths.extend(list(all_simple_paths(graph, missing, target)))
        except (nx.NetworkXNoPath, nx.NodeNotFound):  # noqa: PERF203
            # No path found or nodes not in graph
            continue

    return sorted(paths, key=len)  # Sort by length to show the shortest path first


def _format_paths_msg(nx_graph: Any, paths: list[list[Any]]) -> str:
    """Format paths into a readable message.

    Parameters
    ----------
    paths:
        List of paths, where each path is a list of nodes

    Returns
    -------
    :
        The formatted message
    """
    msg = f"Missing input node '{key_name(paths[0][0])}'. "
    msg += "Affects requested targets (via providers given in parentheses):"

    for i, path in enumerate(paths):
        msg += f"\n{i + 1}. {key_name(path[0])} → "
        msg += " → ".join(
            f"({nx_graph.nodes[node]['provider'].location.qualname}) → {key_name(node)}"
            for node in path[1:]
        )

    return msg


class Pipeline(DataGraph):
    """A container for providers that can be assembled into a task graph."""

    def __init__(
        self,
        providers: Iterable[ToProvider | Provider] | None = None,
        *,
        params: dict[type[Any], Any] | None = None,
        constraints: Mapping[TypeVar, Iterable[Key]] | None = None,
    ):
        """
        Setup a Pipeline from a list providers

        Parameters
        ----------
        providers:
            List of callable providers. Each provides its return value.
            Their arguments and return value must be annotated with type hints.
        params:
            Dictionary of concrete values to provide for types.
        constraints:
            Mapping of type variables to constraints for those type variables.
            For each entry, the corresponding type variable will be constrained
            to the given types.
            This overrides the constraints from the definition of the type variable.
            The new constraints must be a subset of the constraints in
            the type variable definition.
        """
        super().__init__(providers, constraints=constraints)
        for tp, param in (params or {}).items():
            self[tp] = param

    @overload
    def compute(
        self, tp: type[T], reporter: Reporter | None = None, **kwargs: Any
    ) -> T: ...

    @overload
    def compute(
        self, tp: Iterable[type[T]], reporter: Reporter | None = None, **kwargs: Any
    ) -> dict[type[T], T]: ...

    @overload
    def compute(
        self, tp: UnionType, reporter: Reporter | None = None, **kwargs: Any
    ) -> Any: ...

    def compute(
        self,
        tp: type | Iterable[type] | "UnionType",  # noqa: UP037 (needed by Sphinx)
        reporter: Reporter | None = None,
        **kwargs: Any,
    ) -> Any:
        """
        Compute result for the given keys.

        Equivalent to ``self.get(tp).compute()``.

        Parameters
        ----------
        tp:
            Type to compute the result for.
            Can be a single type or an iterable of types.
        reporter:
            Optional reporter to track progress of this computation.
        kwargs:
            Keyword arguments passed to the ``.get()`` method.
        """
        return self.get(tp, **kwargs).compute(reporter=reporter)

    def visualize(
        self,
        tp: type | Iterable[type] | None = None,
        compact: bool = False,
        mode: Literal['data', 'task', 'both'] = 'data',
        cluster_generics: bool = True,
        cluster_color: str | None = '#f0f0ff',
        **kwargs: Any,
    ) -> graphviz.Digraph:
        """
        Return a graphviz Digraph object representing the graph for the given keys.

        Equivalent to ``self.get(tp).visualize()``.

        Parameters
        ----------
        tp:
            Type to visualize the graph for.
            Can be a single type or an iterable of types.
        compact:
            If True, parameter-table-dependent branches are collapsed into a single copy
            of the branch. Recommended for large graphs with long parameter tables.
        mode:
            If 'data', only data nodes are shown. If 'task', only task nodes and input
            data nodes are shown. If 'both', all nodes are shown.
        cluster_generics:
            If True, generic products are grouped into clusters.
        cluster_color:
            Background color of clusters. If None, clusters are dotted.
        kwargs:
            Keyword arguments passed to :py:class:`graphviz.Digraph`.
        """
        if tp is None:
            tp = self.output_keys()
        return self.get(tp, handler=HandleAsComputeTimeException()).visualize(
            compact=compact,
            mode=mode,
            cluster_generics=cluster_generics,
            cluster_color=cluster_color,
            **kwargs,
        )

    def get(
        self,
        keys: type | Iterable[type] | "UnionType",  # noqa: UP037 (needed by Sphinx)
        *,
        scheduler: Scheduler | None = None,
        handler: ErrorHandler | None = None,
        max_depth: int = 4,
    ) -> TaskGraph:
        """
        Return a TaskGraph for the given keys.

        Parameters
        ----------
        keys:
            Type to compute the result for.
            Can be a single type or an iterable of types.
        scheduler:
            Optional scheduler to use for computing the result. If not given, a
            :py:class:`NaiveScheduler` is used if `dask` is not installed,
            otherwise dask's threaded scheduler is used.
        handler:
            Handler for unsatisfied requirements. If not provided,
            :py:class:`HandleAsBuildTimeException` is used, which raises an exception.
            During development and debugging it can be helpful to use a handler that
            raises an exception only when the graph is computed. This can be achieved
            by passing :py:class:`HandleAsComputeTimeException` as the handler.
        max_depth:
            Maximum depth to show in the dependency tree when reporting errors.
        """
        if multi := _is_multiple_keys(keys):
            targets = tuple(keys)  # type: ignore[arg-type]
        else:
            targets = (keys,)  # type: ignore[assignment]
        try:
            graph = to_task_graph(self, targets=targets, handler=handler)
        except UnsatisfiedRequirement as e:
            missing = e.args[1]
            nx_graph = self.underlying_graph
            if missing in nx_graph:
                paths = _find_paths_to_targets(nx_graph, missing, targets)
                info = _format_paths_msg(nx_graph, paths)
            else:
                nodes = ", ".join(map(key_name, nx_graph.nodes))
                info = f'{e} Requested node not in graph. Did you mean one of: {nodes}?'
            # Not raising `from e` because that includes noisy traceback of internals,
            # which are not relevant to the user.
            raise type(e)(f'{info}\n\n') from None
        return TaskGraph(
            graph=graph,
            targets=targets if multi else keys,  # type: ignore[arg-type]
            scheduler=scheduler,
        )

    @overload
    def bind_and_call(self, fns: Callable[..., T], /) -> T: ...

    @overload
    def bind_and_call(
        self, fns: Iterable[Callable[..., Any]], /
    ) -> tuple[Any, ...]: ...

    def bind_and_call(
        self, fns: Callable[..., Any] | Iterable[Callable[..., Any]], /
    ) -> Any:
        """
        Call the given functions with arguments provided by the pipeline.

        Parameters
        ----------
        fns:
            Functions to call.
            The pipeline will provide all arguments based on the function's type hints.

            If this is a single callable, it is called directly.
            Otherwise, ``bind_and_call`` will iterate over it and call all functions.
            If will in either case call :meth:`Pipeline.compute` only once.

        Returns
        -------
        :
            The return values of the functions in the same order as the functions.
            If only one function is passed, its return value
            is *not* wrapped in a tuple.
        """
        return_tuple = True
        if callable(fns):
            fns = (fns,)
            return_tuple = False

        arg_types_per_function = {
            fn: {
                name: ty for name, ty in get_type_hints(fn).items() if name != 'return'
            }
            for fn in fns
        }
        all_arg_types = tuple(
            set(chain(*(a.values() for a in arg_types_per_function.values())))
        )
        values_per_type = self.compute(all_arg_types)
        results = tuple(
            fn(**{name: values_per_type[ty] for name, ty in arg_types.items()})
            for fn, arg_types in arg_types_per_function.items()
        )
        if not return_tuple:
            return results[0]
        return results

    def _repr_html_(self) -> str:
        nodes = ((key, data) for key, data in self.underlying_graph.nodes.items())
        return pipeline_html_repr(nodes)

    def output_keys(self) -> tuple[Key, ...]:
        """Returns the keys that are not inputs to any other providers."""
        sink_nodes = [
            node for node, degree in self.underlying_graph.out_degree if degree == 0
        ]
        return tuple(sorted(sink_nodes, key=key_name))


def _import_pandas(function_name: str) -> pandas:
    try:
        import pandas as pd

        return pd
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"{function_name} requires Pandas. Install Pandas using, e.g., "
            f"`pip install pandas` or `conda install pandas`."
        ) from None


def get_mapped_node_names(
    graph: DataGraph, base_name: type, *, index_names: Sequence[Hashable] | None = None
) -> pandas.Series:
    """
    Given a graph with a mapped node with given base_name, return a series of
    corresponding mapped names.

    This is meant to be used in combination with :py:func:`DataGraph.map`.
    If the mapped node depends on multiple indices, the index of the returned series
    will have a multi-index.

    Attention
    ---------
    This function requires `Pandas <https://pandas.pydata.org/>`_
    which is not installed automatically as a dependency of Sciline.

    Parameters
    ----------
    graph:
        The data graph to get the mapped node names from.
    base_name:
        The base name of the mapped node to get the names for.
    index_names:
        Specifies the names of the indices of the mapped node. If not given this is
        inferred from the graph, but the argument may be required to disambiguate
        multiple mapped nodes with the same name.

    Returns
    -------
    :
        The series of node names corresponding to the mapped node.
    """
    pd = _import_pandas("sciline.get_mapped_node_names")
    from cyclebane.graph import IndexValues, MappedNode, NodeName

    candidates = [
        node
        for node in graph.underlying_graph.nodes
        if isinstance(node, MappedNode) and node.name == base_name
    ]
    if len(candidates) == 0:
        raise ValueError(f"'{base_name}' is not a mapped node.")
    if index_names is not None:
        candidates = [
            node for node in candidates if set(node.indices) == set(index_names)
        ]
    if len(candidates) > 1:
        raise ValueError(
            f"Multiple mapped nodes with name '{base_name}' found: {candidates}"
        )

    index_names = tuple(reversed(candidates[0].indices))
    indices = {name: idx for name, idx in graph.indices.items() if name in index_names}

    index = pd.MultiIndex.from_product(indices.values(), names=index_names)
    keys = tuple(NodeName(base_name, IndexValues(index_names, idx)) for idx in index)
    if index.nlevels == 1:  # Avoid more complicated MultiIndex if unnecessary
        index = index.get_level_values(0)
    return pd.Series(keys, index=index, name=base_name)


def compute_mapped(
    pipeline: Pipeline,
    base_name: type,
    *,
    index_names: Sequence[Hashable] | None = None,
) -> pandas.Series:
    """
    Given a graph with a mapped node with given base_name, return a series of computed
    results.

    This is meant to be used in combination with :py:func:`Pipeline.map`.
    If the mapped node depends on multiple indices, the index of the returned series
    will have a multi-index.

    Attention
    ---------
    This function requires `Pandas <https://pandas.pydata.org/>`_
    which is not installed automatically as a dependency of Sciline.

    Parameters
    ----------
    pipeline:
        The data graph to get the mapped node names from.
    base_name:
        The base name of the mapped node to get the names for.
    index_names:
        Specifies the names of the indices of the mapped node. If not given this is
        inferred from the graph, but the argument may be required to disambiguate
        multiple mapped nodes with the same name.

    Returns
    -------
    :
        The series of computed results corresponding to the mapped node.
    """
    _ = _import_pandas("sciline.compute_mapped")  # required by get_mapped_node_names
    key_series = get_mapped_node_names(
        graph=pipeline, base_name=base_name, index_names=index_names
    )
    results = pipeline.compute(key_series)
    return key_series.apply(lambda x: results[x])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""Progress reporters for pipelines."""

from __future__ import annotations

import functools
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable
from time import perf_counter
from types import TracebackType
from typing import TYPE_CHECKING, Any

from ._provider import Provider
from ._utils import provider_full_qualname, provider_name

if TYPE_CHECKING:
    try:
        import pandas as pd
    except ModuleNotFoundError:
        pd = object


class Reporter(ABC):
    """Base class for progress reporters of computations.

    A ``Reporter`` instance must not be used for multiple computations concurrently!

    Reporters should be passed to :meth:`sciline.Pipeline.compute`, e.g.:

    .. code-block:: python

         from sciline import Pipeline
         from sciline.reporter import RichReporter
         reporter = RichReporter()
         pipeline = Pipeline(providers)
         pipeline.compute(Result, reporter=reporter)
    """

    def __init__(self) -> None:
        self._n_steps = 0
        self._provider_id = -1

    def run_computation(self, providers: Iterable[Provider]) -> Reporter:
        """Run a single computation with progress reporting.

        This method is primarily meant for internal use.
        As a user, you should pass reporters to :meth:`Pipeline.compute`.

        Use this method in a ``with`` statement:

        .. code-block:: python

            with reporter.run_computation(providers):
                # Perform computation

        This method exists mainly to provide arguments when entering a reporting
        context, implementers should mainly override ``__enter__``.
        """
        self._n_steps = sum(1 for provider in providers if provider.kind == "function")
        return self

    @abstractmethod
    def __enter__(self) -> None:
        """Computation is starting."""

    @abstractmethod
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Computation is finished."""

    @abstractmethod
    def on_provider_start(self, provider: Provider) -> int:
        """Called when a provider is started."""

    @abstractmethod
    def on_provider_end(self, provider_id: int) -> None:
        """Called when a provider has finished."""

    def reporting_provider_func(self, provider: Provider) -> Callable[..., Any]:
        """Wrap a provider's function to report progress with this reporter."""
        if provider.kind != "function":
            return provider.func

        @functools.wraps(provider.func)
        def reporting_func(*args: Any, **kwargs: Any) -> Any:
            provider_id = self.on_provider_start(provider)
            try:
                result = provider.func(*args, **kwargs)
            finally:
                self.on_provider_end(provider_id)
            return result

        return reporting_func

    def call_provider_with_reporting(
        self, provider: Provider, values: dict[Hashable, Any]
    ) -> Any:
        """Call a provider and report its progress with this reporter."""
        if provider.kind != "function":
            return provider.call(values)

        provider_id = self.on_provider_start(provider)
        try:
            result = provider.call(values)
        finally:
            self.on_provider_end(provider_id)
        return result

    def _get_provider_id(self) -> int:
        """Return a unique identifier for a current provider."""
        self._provider_id += 1
        return self._provider_id


class TimingReporter(Reporter):
    """A reporter that tracks the time spent in each provider.

    Examples
    --------

    .. code-block:: python

        from sciline.reporter import TimingReporter
        timer = TimingReporter()
        pipeline.compute(Target, reporter=timer)
        print(timer.summary())
    """

    def __init__(self) -> None:
        super().__init__()
        self._times: dict[int, list[float | str]] = {}
        self._total_start = 0.0
        self._total_end = 0.0

    def __enter__(self) -> None:
        """Start the clock for the entire computation."""
        self._times = {}
        self._total_start = perf_counter()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Stop the clock for the entire computation."""
        self._total_end = perf_counter()

    def on_provider_start(self, provider: Provider) -> int:
        """Start the clock for a provider call."""
        provider_id = super()._get_provider_id()
        self._times[provider_id] = [
            perf_counter(),
            0.0,
            provider_full_qualname(provider),
        ]
        return provider_id

    def on_provider_end(self, provider_id: int) -> None:
        """Stop the clock for a provider call."""
        self._times[provider_id][1] = perf_counter()

    def summary(self) -> str:
        """Return a summary of the recorded times.

        Returns
        -------
        :
            The summary as a string.

        See Also
        --------
        TimingReporter.as_pandas:
            Return the times as a Pandas DataFrame.
        """
        sorted_times = self._sorted_times()

        names = ['Provider', *(name for name, _ in sorted_times)]
        lens = _make_column(['N', *(f'({len(times)})' for _, times in sorted_times)])
        sums = _make_column(
            ['Sum [ms]', *(f'{sum(times):.3f}' for _, times in sorted_times)]
        )
        means = _make_column(
            [
                'Mean [ms]',
                *(f'{sum(times) / len(times):.3f}' for _, times in sorted_times),
            ]
        )

        return (
            f'Total time: {self._total_end - self._total_start:.3f} ms\n'
            + "\n".join(
                f'{s} {m} {ll} {n}'
                for s, m, ll, n in zip(sums, means, lens, names, strict=True)
            )
        )

    def as_pandas(self) -> pd.DataFrame:
        """Return the recorded times as a Pandas DataFrame.

        See Also
        --------
        TimingReporter.summary:
            Return a summary of the recorded times as a string.
            Does not require Pandas.
        """
        import numpy as np  # pandas depends on numpy
        import pandas as pd

        return pd.DataFrame(
            [
                (
                    name,
                    len(times),
                    np.sum(times),
                    np.min(times),
                    np.max(times),
                    np.median(times),
                    np.mean(times),
                    np.std(times),
                )
                for name, times in self._sorted_times()
            ],
            columns=[
                'Provider',
                'N Runs',
                'Sum [ms]',
                'Min [ms]',
                'Max [ms]',
                'Median [ms]',
                'Mean [ms]',
                'Std [ms]',
            ],
        )

    def _sorted_times(self) -> list[tuple[str, list[float]]]:
        """Return the recorded times in milliseconds as a sorted list."""
        run_times: dict[str, list[float]] = {}
        for start, end, name in self._times.values():
            run_times.setdefault(name, []).append((end - start) * 1000)  # type: ignore[arg-type, operator]
        return sorted(run_times.items(), key=lambda x: sum(x[1]), reverse=True)


def _make_column(values: list[str]) -> list[str]:
    width = max(len(v) for v in values)
    return [f'{v:>{width}}' for v in values]


class RichReporter(Reporter):
    """Report progress using Rich.

    This class uses the `rich <https://rich.readthedocs.io/en/stable/index.html>`_
    package to display a progress bar.
    Rich is not a hard dependency of Sciline and must be installed separately.

    Examples
    --------

    .. code-block:: python

         from sciline import Pipeline
         from sciline.reporter import RichReporter
         reporter = RichReporter()
         pipeline = Pipeline(providers)
         pipeline.compute(Result, reporter=reporter)
    """

    def __init__(
        self,
        *,
        description: str = "Computing",
        show_remaining_time: bool = False,
        **rich_progress_args: Any,
    ) -> None:
        """Initialize a RichReporter.

        Parameters
        ----------
        description:
            Description shown with the progress bar.
        show_remaining_time:
            Show an estimate of the remaining time.
            This is off by default because providers tend to have very different
            running times and the estimate is unlikely to be useful.
        rich_progress_args:
            Additional arguments passed to the
            `rich.Progress <https://rich.readthedocs.io/en/stable/reference/progress.html#rich.progress.Progress>`_
            constructor.
        """

        from rich import progress

        super().__init__()
        self._description = description
        self._show_remaining_time = show_remaining_time
        self._progress_args = rich_progress_args

        self._active_provider_list = _ProviderList()
        self._progress_bar = _make_rich_progress_bar(
            active_provider_list=self._active_provider_list,
            show_remaining_time=show_remaining_time,
            **rich_progress_args,
        )
        self._task_id = progress.TaskID(0)

    def __enter__(self) -> None:
        """Computation is starting."""
        self._task_id = self._progress_bar.add_task(
            self._description, total=self._n_steps
        )
        self._progress_bar.__enter__()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Computation is finished."""
        self._progress_bar.__exit__(exc_type, exc_val, exc_tb)

    def on_provider_start(self, provider: Provider) -> int:
        """Show that a provider has started."""
        provider_id = self._get_provider_id()
        self._active_provider_list.add_provider(provider_id, provider)
        return provider_id

    def on_provider_end(self, provider_id: int) -> None:
        """Show that a provider has finished."""
        self._active_provider_list.remove_provider(provider_id)
        self._progress_bar.update(self._task_id, advance=1)


def _make_rich_progress_bar(
    active_provider_list: _ProviderList,
    show_remaining_time: bool = False,
    **rich_progress_args: Any,
) -> Any:
    from rich import progress

    columns = [
        progress.TextColumn("[progress.description]{task.description}"),
        progress.BarColumn(),
        progress.MofNCompleteColumn(),
        progress.TimeElapsedColumn(),
    ]
    if show_remaining_time:
        columns.extend(
            (
                progress.RenderableColumn("|"),
                progress.TimeRemainingColumn(),
            )
        )
    columns.append(progress.RenderableColumn(active_provider_list))
    progress_bar = progress.Progress(*columns, **rich_progress_args)
    return progress_bar


class NullReporter(Reporter):
    """A progress reporter that does nothing.

    Useful as a fallback to disable reporting.
    """

    def __enter__(self) -> None:
        """Computation is starting."""

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Computation is finished."""

    def on_provider_start(self, provider: Provider) -> int:
        """Show that a provider has started."""
        return self._get_provider_id()

    def on_provider_end(self, provider_id: int) -> None:
        """Show that a provider has finished."""

    def reporting_provider_func(self, provider: Provider) -> Callable[..., Any]:
        """Call a provider and report its progress with this reporter."""
        # Override base method to avoid overhead.
        return provider.func

    def call_provider_with_reporting(
        self, provider: Provider, values: dict[Hashable, Any]
    ) -> Any:
        """Call a provider and report its progress with this reporter."""
        # Override base method to avoid overhead.
        return provider.call(values)


class _ProviderList:
    """Track the names of currently active providers.

    This class is a
    `rich.console.Renderable https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.ConsoleRenderable`
    that displays the short names of providers.
    """

    def __init__(self) -> None:
        self._provider_names: dict[int, str] = {}

    def add_provider(self, provider_id: int, provider: Provider) -> None:
        self._provider_names[provider_id] = provider_name(provider)

    def remove_provider(self, provider_id: int) -> None:
        del self._provider_names[provider_id]

    def __rich_console__(self, console: Any, options: Any) -> Any:
        if self._provider_names:
            active_providers = rf"\[{', '.join(self._provider_names.values())}]"
        else:
            active_providers = ""
        # Not overflowing the line seems more defensive when this code is used
        # in a larger context.
        options.overflow = 'ellipsis'
        options.no_wrap = True
        return console.render(active_providers, options)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from types import TracebackType
from typing import NewType, TypeVar

from sciline import Pipeline, Provider, Scope
from sciline.reporter import Reporter, RichReporter, TimingReporter


class RecordingReporter(Reporter):
    def __init__(self) -> None:
        super().__init__()
        self.active_providers: dict[int, Provider] = {}
        self.finished_providers: dict[int, Provider] = {}
        self.is_active = False
        self.n_steps = -1

    def __enter__(self) -> None:
        self.n_steps = self._n_steps  # copy from parent class
        self.is_active = True

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.is_active = False

    def on_provider_start(self, provider: Provider) -> int:
        provider_id = self._get_provider_id()
        self.active_providers[provider_id] = provider
        return provider_id

    def on_provider_end(self, provider_id: int) -> None:
        self.finished_providers[provider_id] = self.active_providers.pop(provider_id)


A = NewType('A', int)
B = NewType('B', int)
C = NewType('C', int)
D = NewType('D', int)


def f1(a: A) -> B:
    return B(a + 1)


def f2(b: B, a: A) -> C:
    return C(b * a)


def f3(b: B) -> D:
    return D(b + 2)


def merge(*x: int) -> int:
    return sum(x)


def test_reporter_records_all_function_providers() -> None:
    reporter = RecordingReporter()
    pipeline = Pipeline((f1, f2, f3), params={A: 3})
    result = pipeline.compute(C, reporter=reporter)

    assert result == 12
    assert reporter.n_steps == 2
    assert not reporter.is_active
    assert not reporter.active_providers

    finished_provider_names = sorted(
        provider.func.__name__ for provider in reporter.finished_providers.values()
    )
    assert finished_provider_names == ['f1', 'f2']  # f3 was not called


def test_reporter_records_all_function_providers_map_reduce() -> None:
    reporter = RecordingReporter()
    pipeline = Pipeline((f1, f2, f3), params={})
    pipeline[C] = pipeline[C].map({A: [2, 3]}).reduce(func=merge)
    result = pipeline.compute(C, reporter=reporter)

    assert result == 18
    assert reporter.n_steps == 5
    assert not reporter.is_active
    assert not reporter.active_providers

    finished_provider_names = sorted(
        provider.func.__name__ for provider in reporter.finished_providers.values()
    )
    assert finished_provider_names == ['f1', 'f1', 'f2', 'f2', 'merge']


T1 = NewType('T1', int)
T2 = NewType('T2', int)
T = TypeVar('T', T1, T2)


class X(Scope[T, int], int): ...


class Y(Scope[T, int], int): ...


def g1(x: X[T]) -> Y[T]:
    return Y[T](x + 10)


def g2(y1: Y[T1], y2: Y[T2], a: A) -> B:
    return B(y1 + y2 + a)


def test_reporter_records_all_function_providers_generic() -> None:
    reporter = RecordingReporter()
    pipeline = Pipeline((g1, g2, f2), params={A: 3, X[T1]: 2, X[T2]: 4})
    result = pipeline.compute(C, reporter=reporter)

    assert result == 87
    assert reporter.n_steps == 4
    assert not reporter.is_active
    assert not reporter.active_providers

    finished_provider_names = sorted(
        provider.func.__name__ for provider in reporter.finished_providers.values()
    )
    assert finished_provider_names == ['f2', 'g1', 'g1', 'g2']


def test_timing_reporter_tracks_all_calls() -> None:
    reporter = TimingReporter()
    pipeline = Pipeline((f1, f2, f3), params={})
    pipeline[C] = pipeline[C].map({A: [2, 3]}).reduce(func=merge)
    result = pipeline.compute(C, reporter=reporter)

    assert result == 18
    timings = reporter.as_pandas()
    assert len(timings) == 3
    assert (
        timings.loc[timings['Provider'] == 'tests.reporter_test.f1']['N Runs'].iloc[0]
        == 2
    )
    assert (
        timings.loc[timings['Provider'] == 'tests.reporter_test.f2']['N Runs'].iloc[0]
        == 2
    )
    assert (
        timings.loc[timings['Provider'] == 'tests.reporter_test.merge']['N Runs'].iloc[
            0
        ]
        == 1
    )


def test_rich_reporter_does_not_break_pipeline() -> None:
    # It is very difficult to test the output of RichReporter,
    # but we can at least make sure it doesn't change the final result.
    reporter = RichReporter()
    pipeline = Pipeline((f1, f2, f3), params={A: 3})
    result = pipeline.compute(C, reporter=reporter)
    assert result == 12



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
# ruff: noqa: PYI019
from __future__ import annotations

import itertools
from collections.abc import Callable, Generator, Iterable, Mapping
from types import NoneType
from typing import TYPE_CHECKING, Any, TypeVar, get_args

import cyclebane as cb
import networkx as nx
from cyclebane.node_values import IndexName, IndexValue

from ._provider import ArgSpec, Provider, ToProvider, _bind_free_typevars
from ._utils import key_full_qualname
from .handler import ErrorHandler, HandleAsBuildTimeException
from .typing import Graph, Key

if TYPE_CHECKING:
    import graphviz


def _as_graph(key: Key, value: Any) -> cb.Graph:
    """Create a cyclebane.Graph with a single value."""
    graph = nx.DiGraph()
    graph.add_node(key, value=value)
    return cb.Graph(graph)


def _find_all_typevars(t: type | TypeVar) -> set[TypeVar]:
    """Returns the set of all TypeVars in a type expression."""
    if isinstance(t, TypeVar):
        return {t}
    return set(itertools.chain(*map(_find_all_typevars, get_args(t))))


def _get_typevar_constraints(
    t: TypeVar, over_constraints: dict[TypeVar, frozenset[Key]]
) -> frozenset[Key]:
    """Returns the set of constraints of a TypeVar."""
    if (override := over_constraints.get(t, None)) is not None:
        return override
    if not (constraints := t.__constraints__):
        raise ValueError(
            f"Type variable {t!r} has no constraints. Either constrain the type "
            f"variable in its definition or via the 'constraints' argument of Pipeline."
        )
    return frozenset(constraints)


def _mapping_to_constrained(
    type_vars: set[TypeVar], over_constraints: dict[TypeVar, frozenset[Key]]
) -> Generator[dict[TypeVar, Key], None, None]:
    constraints = [_get_typevar_constraints(t, over_constraints) for t in type_vars]
    for combination in itertools.product(*constraints):
        yield dict(zip(type_vars, combination, strict=True))


def _normalize_custom_constraints(
    constraints: Mapping[TypeVar, Iterable[Key]] | None,
) -> dict[TypeVar, frozenset[Key]]:
    if constraints is None:
        return {}

    normalized = {}
    for key, value in constraints.items():
        types = frozenset(value)
        for ty in types:
            if key.__constraints__ and ty not in key.__constraints__:
                raise ValueError(
                    f"Constraint '{key_full_qualname(ty)}' is not valid for type var "
                    f"'{key_full_qualname(key)}' which supports constraints "
                    f"{tuple(map(key_full_qualname, key.__constraints__))}."
                )
        normalized[key] = types
    return normalized


T = TypeVar('T', bound='DataGraph')


class DataGraph:
    def __init__(
        self,
        providers: None | Iterable[ToProvider | Provider],
        *,
        constraints: Mapping[TypeVar, Iterable[Key]] | None = None,
    ) -> None:
        self._constraints = _normalize_custom_constraints(constraints)
        self._cbgraph = cb.Graph(nx.DiGraph())
        for provider in providers or []:
            self.insert(provider)

    @classmethod
    def _from_cyclebane(cls: type[T], graph: cb.Graph) -> T:
        out = cls([])
        out._cbgraph = graph
        return out

    def copy(self: T) -> T:
        cpy = self._from_cyclebane(self._cbgraph.copy())
        cpy._constraints = self._constraints
        return cpy

    def __copy__(self: T) -> T:
        return self.copy()

    @property
    def index_names(self) -> tuple[IndexName, ...]:
        """Names of the indices (dimensions) of the graph."""
        return self._cbgraph.index_names

    @property
    def indices(self) -> dict[IndexName, Iterable[IndexValue]]:
        """Names and values of the indices of the graph."""
        return self._cbgraph.indices

    @property
    def underlying_graph(self) -> nx.DiGraph:
        """The underlying NetworkX graph."""
        return self._cbgraph.graph

    def _get_clean_node(self, key: Key) -> Any:
        """Return node ready for setting value or provider."""
        if key is NoneType:
            raise ValueError('Key must not be None')
        if key in self.underlying_graph:
            self.underlying_graph.remove_edges_from(
                list(self.underlying_graph.in_edges(key))
            )
            self.underlying_graph.nodes[key].pop('value', None)
            self.underlying_graph.nodes[key].pop('provider', None)
            self.underlying_graph.nodes[key].pop('reduce', None)
        else:
            self.underlying_graph.add_node(key)
        return self.underlying_graph.nodes[key]

    def insert(self, provider: ToProvider | Provider, /) -> None:
        """
        Insert a callable into the graph that provides its return value.

        Parameters
        ----------
        provider:
            Either a callable that provides its return value. Its arguments
            and return value must be annotated with type hints.
            Or a ``Provider`` object that has been constructed from such a callable.
        """
        if not isinstance(provider, Provider):
            provider = Provider.from_function(provider)
        return_type = provider.deduce_key()
        if typevars := _find_all_typevars(return_type):
            for bound in _mapping_to_constrained(typevars, self._constraints):
                self.insert(provider.bind_type_vars(bound))
            return
        # Trigger UnboundTypeVar error if any input typevars are not bound
        provider = provider.bind_type_vars({})
        self._get_clean_node(return_type)['provider'] = provider
        for dep in provider.arg_spec.keys():
            self.underlying_graph.add_edge(dep, return_type, key=dep)

    def __setitem__(self, key: Key, value: DataGraph | Any) -> None:
        """
        Provide a concrete value for a type.

        Parameters
        ----------
        key:
            Type to provide a value for.
        value:
            Concrete value to provide.
        """
        # This is a questionable approach: Using MyGeneric[T] as a key will actually
        # not pass mypy [valid-type] checks. What we do on our side is ok, but the
        # calling code is not.
        if typevars := _find_all_typevars(key):
            for bound in _mapping_to_constrained(typevars, self._constraints):
                self[_bind_free_typevars(key, bound)] = value
            return

        # TODO If key is generic, should we support multi-sink case and update all?
        # Would imply that we need the same for __getitem__.
        self._cbgraph[key] = (
            value._cbgraph if isinstance(value, DataGraph) else _as_graph(key, value)
        )

    def __getitem__(self: T, key: Key) -> T:
        """Return the subgraph that computes the given key."""
        return self._from_cyclebane(self._cbgraph[key])

    def map(self: T, node_values: dict[Key, Any]) -> T:
        """Map the graph over given node values.

        Creates a new graph where given nodes and their dependents are duplicated for
        each given value and values are assigned to the given nodes.

        Parameters
        ----------
        node_values:
            Dictionary mapping nodes keys to collections of values.

        Returns
        -------
        :
            A new graph with mapped nodes.
        """
        return self._from_cyclebane(self._cbgraph.map(node_values))

    def reduce(self: T, *, func: Callable[..., Any], **kwargs: Any) -> T:
        """Reduce the outputs of a mapped graph into a single value and provider.

        Parameters
        ----------
        func:
            Function that takes the values to reduce and returns a single value.
            This function is passed as many arguments as there are values to reduce.
        kwargs:
            Forwarded to :meth:`cyclebane.Graph.reduce`.

        Returns
        -------
        :
            A new graph with a new node that depends on all sink nodes of the input
            graph and returns the output of ``func``.
        """
        # Note that the type hints of `func` are not checked here. As we are explicit
        # about the modification, this is in line with __setitem__ which does not
        # perform such checks and allows for using generic reduction functions.
        return self._from_cyclebane(
            self._cbgraph.reduce(attrs={'reduce': func}, **kwargs)
        )

    def to_networkx(self) -> nx.DiGraph:
        return self._cbgraph.to_networkx()

    def visualize_data_graph(self, **kwargs: Any) -> graphviz.Digraph:
        import graphviz

        dot = graphviz.Digraph(strict=True, **kwargs)
        for node in self.underlying_graph.nodes:
            dot.node(str(node), label=str(node), shape='box')
            attrs = self.underlying_graph.nodes[node]
            attrs = '\n'.join(f'{k}={v}' for k, v in attrs.items())
            dot.node(str(node), label=f'{node}\n{attrs}', shape='box')
        for edge in self.underlying_graph.edges:
            key = self.underlying_graph.edges[edge].get('key')
            label = str(key) if key is not None else ''
            dot.edge(str(edge[0]), str(edge[1]), label=label)
        return dot


_no_value = object()


def to_task_graph(
    data_graph: DataGraph, targets: tuple[Key, ...], handler: ErrorHandler | None = None
) -> Graph:
    graph = data_graph.to_networkx()
    handler = handler or HandleAsBuildTimeException()
    ancestors = list(targets)
    for target in targets:
        if target not in graph:
            handler.handle_unsatisfied_requirement(target)
        ancestors.extend(nx.ancestors(graph, target))
    graph = graph.subgraph(set(ancestors))
    out = {}

    for key in graph.nodes:
        node = graph.nodes[key]
        input_nodes = list(graph.predecessors(key))
        input_edges = list(graph.in_edges(key, data=True))
        orig_keys = [edge[2].get('key', None) for edge in input_edges]
        if (value := node.get('value', _no_value)) is not _no_value:
            out[key] = Provider.parameter(value)
        elif (provider := node.get('provider')) is not None:
            new_key = dict(zip(orig_keys, input_nodes, strict=True))
            # By using map_keys (instead of creating an ArgSpec from scratch),
            # we automatically preserve what args and kwargs are.
            spec = provider.arg_spec.map_keys(new_key.get)
            if len(spec) != len(input_nodes):
                # This should be caught by __setitem__, but we check here to be safe.
                raise ValueError("Corrupted graph")
            out[key] = Provider(func=provider.func, arg_spec=spec, kind='function')
        elif (func := node.get('reduce')) is not None:
            spec = ArgSpec.from_args(*input_nodes)
            out[key] = Provider(func=func, arg_spec=spec, kind='function')
        else:
            out[key] = handler.handle_unsatisfied_requirement(key)
    return out
